'use strict';

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 * How it works:
 * 1. It makes initialization - gets config, called template factories, sets flag to default state.
 * 2. Plugin traverse in 3 ways:
 *      - basic function declarations (FunctionDeclaration, ArrowFunctionExpression, etc.)
 *      - variable declarations with declared function
 *      - return statement with function as argument
 * 3. It tries to find closest block comment, that can be relative to this function.
 * 4. It tries to parse it with doctrine (https://github.com/eslint/doctrine).
 * 5. It makes function body normalization (removes "(a) => a * a" cases, makes it multiline)
 * 6. It inserts type check function call for arguments.
 * 7. It makes traverse for all return statements, relative to this function.
 * 8. It adds type check function call into return statement.
 * 9. If some function were transformed by plugin, on exit of Program it adds type check function declaration to file.
 */

var config = require('../config.json');
var findGlobalDirective = require('./lib/find-global-directive');
var checkFunction = require('./lib/check-function');
var findComment = require('./lib/find-comment');
var parseJsDoc = require('./lib/parse-jsdoc');
var normalizeFunctionBody = require('./lib/normalize-function-body');
var insertParametersCheck = require('./lib/insert-parameters-check');
var normalizeValidator = require('./lib/normalize-validator');
var typecheckTemplate = require('./lib/typecheck-function-template');

var VISITORS = ['ClassMethod', 'ObjectMethod', 'FunctionDeclaration', 'ArrowFunctionExpression'];

/**
 * @param {Object} t
 * @returns {{visitor: Object}}
 */
module.exports = function (_ref) {
    var t = _ref.types;

    var typecheckFunctionDeclaration = typecheckTemplate.function(config.functionName);
    var typecheckFunctionCall = typecheckTemplate.call(config.functionName, t);

    var hasGlobalDirective = false;
    var shouldInjectHelperFunction = false;

    var HELPERS = {
        isDeclarationIsFunction: function isDeclarationIsFunction(declaration) {
            return declaration.get('init').isFunction();
        },
        isPathIsConstructor: function isPathIsConstructor(path) {
            return path.node.kind === 'constructor';
        },
        isPathIsExpression: function isPathIsExpression(path) {
            return path.isExpressionStatement();
        }
    };

    /**
     * @param {String} comment
     * @param {NodePath} path
     * @param {String} [name]
     */
    function executeFunctionTransformation(comment, path, name) {
        var jsDoc = parseJsDoc(comment);

        if (!jsDoc) {
            return;
        }

        var functionName = name || jsDoc.name || (path.node.id ? path.node.id.name : config.defaultFunctionName);

        normalizeFunctionBody(path);

        if (!checkFunction(path)) {
            return;
        }

        insertParametersCheck(functionName, typecheckFunctionCall, path, jsDoc, t);

        if (jsDoc.returnStatement) {
            path.traverse(returnTypecheckVisitor, {
                jsDoc: jsDoc,
                functionPath: path,
                functionName: functionName
            });
        }

        shouldInjectHelperFunction = true;
    }

    var returnTypecheckVisitor = {
        ReturnStatement: function ReturnStatement(path) {
            if (path.getFunctionParent().node !== this.functionPath.node) {
                return;
            }

            var statement = this.jsDoc.returnStatement;
            var argument = path.get('argument');

            var functionCall = typecheckFunctionCall(this.functionName, 'return', argument.node || t.identifier('void 0'), normalizeValidator(path, statement, t));

            argument.replaceWith(functionCall.expression);
        }
    };

    return {
        visitor: _defineProperty({
            Program: {
                enter: function enter(path, state) {
                    hasGlobalDirective = findGlobalDirective(path, state);
                    shouldInjectHelperFunction = false;
                },
                exit: function exit(path, state) {
                    if (!shouldInjectHelperFunction || state.opts.insertHelper === false) {
                        return;
                    }

                    path.pushContainer('body', typecheckFunctionDeclaration());
                }
            },

            ClassDeclaration: function ClassDeclaration(path, state) {
                var classBody = path.get('body');
                var nodes = classBody.get('body');
                var constructor = nodes.find(HELPERS.isPathIsConstructor);

                if (!constructor) {
                    return;
                }

                var comment = findComment(constructor, state, hasGlobalDirective);

                if (!comment) {
                    return;
                }

                executeFunctionTransformation(comment, constructor, path.node.id.name + '.constructor');
            },


            /**
             * @param {NodePath} path
             * @param {PluginPass} state
             */
            VariableDeclaration: function VariableDeclaration(path, state) {
                var comment = findComment(path, state, hasGlobalDirective);

                if (!comment) {
                    return;
                }

                var declarations = path.get('declarations');
                var functionDeclaration = declarations.find(HELPERS.isDeclarationIsFunction);

                if (!functionDeclaration) {
                    return;
                }

                executeFunctionTransformation(comment, functionDeclaration.get('init'));
            },
            ReturnStatement: function ReturnStatement(path, state) {
                var argument = path.get('argument');

                if (!argument || !argument.isFunction()) {
                    return;
                }

                var comment = findComment(path, state, hasGlobalDirective);

                if (!comment) {
                    return;
                }

                executeFunctionTransformation(comment, argument);
            },
            AssignmentExpression: function AssignmentExpression(path, state) {
                var rightPath = path.get('right');

                if (!rightPath.isFunction()) {
                    return;
                }

                var expression = path.find(HELPERS.isPathIsExpression);
                var comment = findComment(expression, state, hasGlobalDirective);

                if (!comment) {
                    return;
                }

                executeFunctionTransformation(comment, rightPath);
            },
            ObjectProperty: function ObjectProperty(path, state) {
                var value = path.get('value');

                if (!value.isFunction()) {
                    return;
                }

                var comment = findComment(path, state, hasGlobalDirective);

                if (!comment) {
                    return;
                }

                executeFunctionTransformation(comment, value);
            }
        }, VISITORS.join('|'), function (path, state) {
            var comment = findComment(path, state, hasGlobalDirective);

            if (!comment) {
                return;
            }

            executeFunctionTransformation(comment, path);
        })
    };
};